import traceback

import tornado.web
from jaeger_client import Span
from opentracing import Format
from opentracing import tags

from py_jaeger_tracing.main import TracingEnvironment


class TracerMixin:
    def __init__(self, *args, **kwargs):
        self._request_span: Span = None

    def initialize_tracer(self, request):
        tracer = TracingEnvironment.tracer

        if tracer:
            parent_span = tracer.extract(Format.HTTP_HEADERS, request.headers)

            self._request_span = tracer.start_span(
                operation_name=f'{request.method} {request.path}',
                child_of=parent_span
            )
            TracingEnvironment.spans.append(self._request_span)

            for k, v in {
                tags.SPAN_KIND: tags.SPAN_KIND_RPC_SERVER,
                tags.HTTP_URL: request.path,
                tags.HTTP_METHOD: request.method,
                tags.PEER_HOST_IPV4: request.remote_ip,
                tags.COMPONENT: 'tornado',
            }.items():
                self._request_span.set_tag(k, v)

    def exception_tracer(self, status_code, **kwargs):
        exception, reason, tb = kwargs['exc_info']

        if self._request_span:
            self._request_span.set_tag(tags.HTTP_STATUS_CODE, status_code)
            self._request_span.set_tag(tags.ERROR, True)
            self._request_span.log_kv({
                'event': tags.ERROR,
                'error.object': exception,
                'error.traceback': traceback.format_tb(tb)
            })
            TracingEnvironment.spans.pop()


class BaseRequestHandler(tornado.web.RequestHandler, TracerMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def data_received(self, chunk):
        pass

    def initialize(self):
        self.initialize_tracer(self.request)

    def on_finish(self):
        if TracingEnvironment.tracer and len(TracingEnvironment.spans) > 0:
            TracingEnvironment.spans.pop()
            self._request_span.finish()

    def write_error(self, status_code, **kwargs):
        exception, reason, tb = kwargs['exc_info']
        self.exception_tracer(status_code, **kwargs)

        if isinstance(exception, ValueError):
            self.set_status(400)
        else:
            self.set_status(500)

        if self.settings.get('debug', False):
            self.write({
                'success': False,
                'response': {
                    'reason': exception.__name__,
                    'traceback': traceback.format_tb(tb)
                }
            })
        else:
            self.write({
                'success': False,
                'response': {
                    'reason': str(reason)
                }
            })

    def on_connection_close(self):
        if self._request_span:
            self._request_span.set_tag(tags.ERROR, True)
            self._request_span.finish()
            TracingEnvironment.spans.pop()



from opentracing import Format

from py_jaeger_tracing.main import TracingEnvironment


class TracingDictInjector:
    def __init__(self, operation='child'):
        self._operation = operation

    def _inject(self, carrier):
        if len(TracingEnvironment.spans) > 0:
            span = TracingEnvironment.tracer.start_span(self._operation, child_of=TracingEnvironment.spans[-1])
        else:
            span = TracingEnvironment.tracer.start_span(self._operation)

        TracingEnvironment.tracer.inject(
            span_context=span.context,
            format=Format.TEXT_MAP,
            carrier=carrier
        )
        span.finish()

    def inject(self, carrier):
        if TracingEnvironment.tracer:
            self._inject(carrier)



import enum

from py_jaeger_tracing.patches.requests import patch_requests
from py_jaeger_tracing.patches.boto3 import patch_boto3


class TracingPatch(enum.Enum):
    REQUESTS = patch_requests
    BOTO_3 = patch_boto3


class TracingPatcher:
    @classmethod
    def apply_patches(cls, patches):
        for patch in patches:
            patch()



from opentracing import Format
from opentracing import Reference
from opentracing import ReferenceType

from py_jaeger_tracing.main import TracingEnvironment


class TracingDictExtractor:
    def __init__(self, operation='child'):
        self._operation = operation

    def _extract(self, carrier):
        span_ctx = TracingEnvironment.tracer.extract(
            format=Format.TEXT_MAP,
            carrier=carrier
        )

        references = [Reference(ReferenceType.FOLLOWS_FROM, span_ctx)]
        if len(TracingEnvironment.spans) > 0:
            references += [Reference(ReferenceType.FOLLOWS_FROM, TracingEnvironment.spans[-1])]

        span = TracingEnvironment.tracer.start_span(
            operation_name=self._operation,
            references=references
        )

        TracingEnvironment.spans.append(span)
        span.finish()

    def extract(self, carrier):
        if TracingEnvironment.tracer:
            self._extract(carrier)



import functools
import traceback

from opentracing import tags

from py_jaeger_tracing.main import TracingEnvironment


def _tracing(f, operation_name, *args, **kwargs):
    f_signature = f'{f.__module__}.{f.__name__}({",".join(["?"] * len(args) + list(kwargs.keys()))})'

    operation_name = f'{f_signature}; {operation_name}'
    if len(TracingEnvironment.spans) > 0:
        span = TracingEnvironment.tracer.start_span(operation_name, child_of=TracingEnvironment.spans[-1])
    else:
        span = TracingEnvironment.tracer.start_span(operation_name)

    TracingEnvironment.spans.append(span)

    try:
        return f(*args, **kwargs)
    except Exception as e:
        span.set_tag(tags.ERROR, True)
        span.log_kv({
            'event': tags.ERROR,
            'error.object': e,
            'error.traceback': traceback.format_exc()
        })
        raise e
    finally:
        h_span = None
        while h_span != span:
            h_span = TracingEnvironment.spans.pop()
        span.finish()


class TracingDecorator:
    def __init__(self, operation_name=None):
        self._operation_name = operation_name or ''

    def __call__(self, f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if TracingEnvironment.tracer:
                return _tracing(f, self._operation_name, *args, **kwargs)
            else:
                return f(*args, **kwargs)

        return wrapper



import copy

from jaeger_client import Span
from opentracing import Format
from opentracing import tags


def patch_requests():
    import requests

    from py_jaeger_tracing.main import TracingEnvironment
    from py_jaeger_tracing.utils.context import tracing_context

    real_impl = requests.request

    def fake_impl(method, url, **kwargs):
        kwargs = copy.deepcopy(kwargs)

        with tracing_context('request'):
            if len(TracingEnvironment.spans) > 0:
                span: Span = TracingEnvironment.spans[-1]
                span.set_tag(tags.HTTP_METHOD, method.upper())
                span.set_tag(tags.HTTP_URL, url)
                span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)

                kwargs['headers'] = kwargs.get('headers', {})
                TracingEnvironment.tracer.inject(span, Format.HTTP_HEADERS, kwargs['headers'])

            return real_impl(method, url, **kwargs)

    requests.api.request = fake_impl



import contextlib
import traceback

from opentracing import tags

from py_jaeger_tracing.main import TracingEnvironment


@contextlib.contextmanager
def tracing_context(name='', *, context_function=None):
    if context_function:
        operation_name = f'{context_function.__module__}.{context_function.__name__}: {name}'
    else:
        operation_name = name

    span = None
    try:
        if len(TracingEnvironment.spans) > 0:
            span = TracingEnvironment.tracer.start_span(operation_name, child_of=TracingEnvironment.spans[-1])
        else:
            span = TracingEnvironment.tracer.start_span(operation_name)

        TracingEnvironment.spans.append(span)
        yield
    except Exception as e:
        span.set_tag(tags.ERROR, True)
        span.log_kv({
            'event': tags.ERROR,
            'error.object': e,
            'error.traceback': traceback.format_exc()
        })
        raise e
    finally:
        h_span = None
        while h_span != span:
            h_span = TracingEnvironment.spans.pop()
        span.finish()



from logging import StreamHandler

from py_jaeger_tracing.main import TracingEnvironment


class TracerHandler(StreamHandler):
    def emit(self, record):
        if TracingEnvironment.tracer and len(TracingEnvironment.spans) > 0:
            span = TracingEnvironment.spans[-1]
            span.log_kv({
                'event': f'logger.{record.levelname}',
                'text': self.format(record)
            })
